using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.ServiceProcess;
using System.Text;
using System.ServiceModel;
using System.Configuration;

namespace System.ServiceProcess
{
    /// <summary>
    /// Abstract class that controls the WCF service behaviour
    /// </summary>
    /// <typeparam name="T">service pattern</typeparam>
    public abstract class CmdLineService<T> : ServiceBase, ICmdLineStarter
    where T : new()
    {
        #region Parameters

        /// <summary>
        /// the class parameters
        /// </summary>
        private class CmdLineServiceParameters
        {
            //Default values
            public const int DEFAULT_RETRYNUMBERAUTHORIZED = 3;
            public const int DEFAULT_LOGLEVELREQUIRED = 1;
            public static string DEFAULT_CATEGORY = String.Empty;

            //Parameters (inner fields)
            private int retryNumberAuthorized;
            private int logLevelRequired;
            private TraceSwitch traceSwitch;
            private string traceCategory;

            /// <summary>
            /// Number of auto-reconnections to WCF before giving up
            /// </summary>
            public int RetryNumberAuthorized
            {
                get { return retryNumberAuthorized; }
                set { retryNumberAuthorized = value; }
            }
            /// <summary>
            /// Log level required
            /// </summary>
            public int LogLevelRequired
            {
                get { return logLevelRequired; }
                set { logLevelRequired = value; }
            }
            /// <summary>
            /// Switch the traces level :
            /// </summary>
            public TraceSwitch TraceSwitch
            {
                get { return traceSwitch; }
                set { traceSwitch = value; }
            }
            /// <summary>
            /// The Trace output Category
            /// </summary>
            public string TraceCategory
            {
                get { return traceCategory; }
                set { traceCategory = value; }
            }

            /// <summary>
            /// Default contructor
            /// </summary>
            public CmdLineServiceParameters()
            {
                UpdateConf();
            }

            /// <summary>
            /// Loading parameters
            /// </summary>
            public void UpdateConf()
            {
                ////////////
                //Allocating default values :
                //
                this.retryNumberAuthorized = DEFAULT_RETRYNUMBERAUTHORIZED;
                this.logLevelRequired = DEFAULT_LOGLEVELREQUIRED;
                this.traceSwitch = new TraceSwitch("CmdLineSwitch", "inner TraceSwitch");
                this.traceSwitch.Level = (TraceLevel) this.logLevelRequired;
                this.traceCategory = DEFAULT_CATEGORY;

                ////////////
                //Reallocating new values :
                //
                String confValue = String.Empty;
                foreach (string aValue in ConfigurationManager.AppSettings)
                {
                    switch (aValue)
                    {
                        case "CmdLineService.retryNumberAuthorized":
                            //Number of connexion retry before giving up.
                            confValue = ConfigurationManager.AppSettings[aValue];
                            if (String.IsNullOrEmpty(confValue) == false)
                            {
                                if (int.TryParse(confValue, out this.retryNumberAuthorized) == false)
                                    this.retryNumberAuthorized = DEFAULT_RETRYNUMBERAUTHORIZED;
                            }
                            break;

                        case "CmdLineService.logLevelRequired":
                            //Log level required
                            confValue = ConfigurationManager.AppSettings[aValue];
                            if (String.IsNullOrEmpty(confValue) == false)
                            {
                                if (int.TryParse(confValue, out this.logLevelRequired) == false)
                                    this.logLevelRequired = DEFAULT_LOGLEVELREQUIRED;
                            }
                            //updating dependences
                            traceSwitch.Level = (TraceLevel) this.logLevelRequired;
                            break;

                        case "CmdLineService.traceCategory":
                            //Log category
                            confValue = ConfigurationManager.AppSettings[aValue];
                            if (String.IsNullOrEmpty(confValue) == false)
                                this.traceCategory = confValue;
                            break;
                    }
                }
            }
        }
        /// <summary>
        /// Local parameters
        /// </summary>
        private CmdLineServiceParameters myParameters = new CmdLineServiceParameters();

        //Internal parameters
        private int nbRestartRequested = 0;
        private DateTime dtLastRetry = DateTime.Now.Date;

        #endregion

        #region Direct access to components

        /// <summary>
        /// Service
        /// </summary>
        ServiceHost m_myServiceHost = null;
        /// <summary>
        /// Object T 
        /// </summary>
        private static T m_refObject = new T();


        /// <summary>
        /// Boxing of the T valriable
        /// </summary>
        public static T RefObject
        {
            get { return CmdLineService<T>.m_refObject; }
        }
        /// <summary>
        /// Holding the service
        /// </summary>
        public ServiceHost MyServiceHost
        {
            get { return m_myServiceHost; }
            set { m_myServiceHost = value; }
        }

        /// <summary>
        /// Gets or defines the service name
        /// </summary>
        new public abstract string ServiceName { get; }

        #endregion

        #region WCF events management

        /// <summary>
        /// Starter function
        /// </summary>
        public virtual void StartHosting()
        {
            //Tracing informations :
            TraceInfo(TraceLevel.Verbose, "StartHosting");

            //Checking host current status
            if ((m_myServiceHost == null) ||
                (m_myServiceHost.State != CommunicationState.Opened))
            {
                //Checking other unrecognized status (like closed or closing) :
                if (m_myServiceHost != null) StopHosting();

                //Startup
                try
                {
                    // Instantiating new ServiceHost 
                    if (Attribute.IsDefined(typeof(T), typeof(ServiceBehaviorAttribute)))
                    {
                        ServiceBehaviorAttribute attrinfo = (ServiceBehaviorAttribute)
                            Attribute.GetCustomAttribute(typeof(T), typeof(ServiceBehaviorAttribute));
                        switch (attrinfo.InstanceContextMode)
                        {
                            case InstanceContextMode.Single:
                                m_myServiceHost = new ServiceHost(m_refObject);
                                break;
                            default:
                                m_myServiceHost = new ServiceHost(typeof(T));
                                break;
                        }
                    }
                    else
                    {
                        m_myServiceHost = new ServiceHost(typeof(T));
                    }
                    m_myServiceHost.Open();
                    m_myServiceHost.Faulted += new EventHandler(m_myServiceHost_Faulted);
                }
                catch (Exception ex)
                {
                    //Setting up null
                    m_myServiceHost = null;
                    if (ex.InnerException != null)
                    //Tracing fault information
                        TraceInfo(TraceLevel.Error, String.Format("Err = {0}\r\nInnerErr = {1}",
                            ex.Message, ex.InnerException.Message));
                    else
                        TraceInfo(TraceLevel.Error, String.Format("Err = {0}",
                            ex.Message));

                    //Trying to start once again...
                    TryAutoRestart();
                }
            }

        }
        /// <summary>
        /// Stopping the host service
        /// </summary>
        public virtual void StopHosting()
        {
            //Tracing informations :
            TraceInfo(TraceLevel.Verbose, "StopHosting");

            //Checking host behaviour
            if (MyServiceHost != null)
            {
                //Forcing host closure :
                if (MyServiceHost.State != CommunicationState.Closed)
                {
                    if (MyServiceHost.State == CommunicationState.Faulted)
                        MyServiceHost.Abort();
                    else
                        MyServiceHost.Close();
                }

                //Backing it up to null
                MyServiceHost = null;
            }
        }
        /// <summary>
        /// Restarting the service (just a stop and start activity)
        /// </summary>
        public virtual void ReStartHosting()
        {
            //Tracing information :
            TraceInfo(TraceLevel.Verbose, "ReStartHosting");

            StopHosting();
            StartHosting();
        }
        /// <summary>
        /// In case of failure, 'Falted' event, restarting the service
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void m_myServiceHost_Faulted(object sender, EventArgs e)
        {
            //Tracing error
            TraceInfo(TraceLevel.Warning, 
                String.Format("Object {0} enterred in Faulted status\r\nerrInfo : {1}",
                m_refObject.ToString(), e.ToString()));

            //Rdmarrage instantan du service
            TryAutoRestart();
        }

        #endregion

        #region Failure and tracing tools

        /// <summary>
        /// This function is used for tracing
        /// This is virtual method just in case you have particular tracing needs
        /// </summary>
        /// <param name="lvlTrace">level of trace concerned</param>
        /// <param name="raisonDuBug">information to log</param>
        public virtual void TraceInfo(TraceLevel lvlTrace, String raisonDuBug)
        {
            //Checking trace requirement :  
            if (myParameters.TraceSwitch.Level > TraceLevel.Off &&
                myParameters.TraceSwitch.Level >= lvlTrace)
            {
                //Building the trace info
                StringBuilder sb = new StringBuilder("CmdLineService : ");
                switch (lvlTrace)
                {
                    case TraceLevel.Error:
                        sb.Append("Error : ");
                        break;
                    case TraceLevel.Info:
                        sb.Append("Information : ");
                        break;
                    case TraceLevel.Verbose:
                        sb.Append("Note : ");
                        break;
                    case TraceLevel.Warning:
                        sb.Append("Warning : ");
                        break;
                    default:
                        break;
                }
                sb.Append(raisonDuBug);

                //Checking the trace category
                if (String.IsNullOrEmpty(myParameters.TraceCategory))
                    Trace.WriteLine(sb);
                else
                    Trace.WriteLine(sb, myParameters.TraceCategory);
            }
        }

        /// <summary>
        /// Auto restart function
        /// </summary>
        private void TryAutoRestart()
        {
            //If the date is older than today,
            //forget the retry-number
            if (this.dtLastRetry < DateTime.Now.Date)
            {
                lock (this)
                {
                    this.dtLastRetry = DateTime.Now.Date;
                    this.nbRestartRequested = 0;
                }
            }

            //Checking the number of retry
            if (this.nbRestartRequested < myParameters.RetryNumberAuthorized)
            {
                //Increment the retry number
                this.nbRestartRequested += 1;

                //Tracing information :
                TraceInfo(TraceLevel.Info, String.Format("restarting WCF service #{0}",
                    this.nbRestartRequested));

                //Restarting
                ReStartHosting();
            }
            else
            {
                //Taux max de redmarrage dpass :
                TraceInfo(TraceLevel.Error, "Max threshold of restarts overcome :\r\n" +
                    "To change it, please modify the 'CmdLineService.retryNumberAuthorized' key " +
                    "in your configuration file");

                //Exporting failure
                //this is because we want the service to fail
                //so that MSCS or other server custering services
                //can detect it to use the passive node instead
                throw new CmdLineException("Cannot reconnect the server, check log for mode details");
            }
        }

        #endregion

        #region ICmdLineStarter Members

        /// <summary>
        /// Manual startup
        /// </summary>
        /// <param name="args"></param>
        public void CmdStart(string[] args)
        {
            TraceInfo(TraceLevel.Info, "Starting the service...");
            OnStart(args);
        }
        /// <summary>
        /// Manual end
        /// </summary>
        public void CmdStop()
        {
            TraceInfo(TraceLevel.Info, "Ending the service...");
            OnStop();
        }

        #endregion

    }

    public class CmdLineException : Exception
    {
        //
        // For guidelines regarding the creation of new exception types, see
        //    http://msdn.microsoft.com/library/default.asp?url=/library/en-us/cpgenref/html/cpconerrorraisinghandlingguidelines.asp
        // and
        //    http://msdn.microsoft.com/library/default.asp?url=/library/en-us/dncscol/html/csharp07192001.asp
        //

        public CmdLineException() { }
        public CmdLineException(string message) : base(message) { }
        public CmdLineException(string message, Exception inner) : base(message, inner) { }
        protected CmdLineException(
          System.Runtime.Serialization.SerializationInfo info,
          System.Runtime.Serialization.StreamingContext context)
            : base(info, context) { }
    };

}
